#
# Author: Facundo Batista <facundo@canonical.com>
#
# Copyright 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
'''Tests the Hashs Queue.'''

from __future__ import with_statement

import os
import unittest
import random
import time

from twisted.trial.unittest import TestCase as TwistedTestCase
from twisted.internet import defer, reactor

from contrib.testing import testcase
from ubuntuone.syncdaemon import hash_queue
from ubuntuone.storageprotocol.hash import \
            content_hash_factory, crc32


# this is a unit test we need to access protected members
# pylint: disable-msg=W0212
class HasherTests(testcase.BaseTwistedTestCase):
    '''Test the whole stuff to receive signals.'''

    def setUp(self):
        '''Setup the test.'''
        testcase.BaseTwistedTestCase.setUp(self)
        self.test_dir = self.mktemp('test_dir')
        self.timeout = 2

    def tearDown(self):
        '''Clean up the tests.'''
        testcase.BaseTwistedTestCase.tearDown(self)

    def test_live_process(self):
        '''Check that the hasher lives and dies.'''
        # create the hasher
        mark = object()
        queue = hash_queue.UniqueQueue()
        class Helper(object):
            '''helper class'''
            def push(self, *args):
                '''callback'''
        receiver = Helper()
        hasher = hash_queue._Hasher(queue, mark, receiver)
        hasher.start()

        # it's aliveeeeeeee!
        self.assertTrue(hasher.isAlive())

        # stop it, and release the processor to let the other thread run
        queue.put(mark)
        time.sleep(.1)

        # "I see dead threads"
        self.assertFalse(hasher.isAlive())


    def test_called_back(self):
        '''Tests that the hasher produces correct info.'''
        # create the hasher
        mark = object()
        queue = hash_queue.UniqueQueue()
        d = defer.Deferred()
        class Helper(object):
            '''helper class'''
            def push(self, *args):
                '''callback'''
                d.callback(args)
        receiver = Helper()
        hasher = hash_queue._Hasher(queue, mark, receiver)
        hasher.start()

        # send what to hash
        testfile = os.path.join(self.test_dir, "testfile")
        with open(testfile, "w") as fh:
            fh.write("foobar")
        queue.put(testfile)

        def check_info(args):
            """check the info pushed by the hasher"""
            # pylint: disable-msg=W0612
            event, path, hash, crc, size, stat = args
            # calculate what we should receive
            realh = content_hash_factory()
            realh.hash_object.update("foobar")
            should_be = realh.content_hash()
            curr_stat = os.stat(testfile)
            self.assertEquals(should_be, hash)
            for attr in ('st_mode', 'st_ino', 'st_dev', 'st_nlink', 'st_uid',
                         'st_gid', 'st_size', 'st_ctime', 'st_mtime'):
                self.assertEquals(getattr(curr_stat, attr),
                                  getattr(stat, attr))

        d.addCallback(check_info)
        # release the processor and check
        return d

    def test_order(self):
        '''The hasher should return in order.'''
        # calculate what we should receive
        should_be = []
        for i in range(10):
            hasher = content_hash_factory()
            text = "supercalifragilistico"+str(i)
            hasher.hash_object.update(text)
            tfile = os.path.join(self.test_dir, "tfile"+str(i))
            with open(tfile, "w") as fh:
                fh.write("supercalifragilistico"+str(i))
            should_be.append(("HQ_HASH_NEW", tfile, hasher.content_hash(),
                              crc32(text), len(text), os.stat(tfile)))

        # create the hasher
        mark = object()
        queue = hash_queue.UniqueQueue()
        d = defer.Deferred()

        class Helper(object):
            '''helper class'''
            # class-closure, cannot use self, pylint: disable-msg=E0213
            def __init__(innerself):
                innerself.store = []
            def push(innerself, *args):
                '''callback'''
                innerself.store.append(args)
                if len(innerself.store) == 10:
                    if innerself.store == should_be:
                        d.callback(True)
                    else:
                        d.errback(Exception("are different!"))
        receiver = Helper()

        hasher = hash_queue._Hasher(queue, mark, receiver)
        hasher.start()

        # send what to hash
        for i in range(10):
            tfile = os.path.join(self.test_dir, "tfile"+str(i))
            queue.put(tfile)
        return d

    def test_large_content(self):
        '''The hasher works ok for a lot of info.'''
        # calculate what we should receive
        testinfo = "".join(chr(random.randint(0, 255)) for i in range(100000))
        hasher = content_hash_factory()
        hasher.hash_object.update(testinfo)
        testfile = os.path.join(self.test_dir, "testfile")
        testhash = hasher.content_hash()

        # create the hasher
        mark = object()
        queue = hash_queue.UniqueQueue()
        d = defer.Deferred()

        class Helper(object):
            '''helper class'''
            def push(self, event, path, newhash, crc, size, stat):
                '''callback'''
                if event != "HQ_HASH_NEW":
                    d.errback(Exception("envent is not HQ_HASH_NEW"))
                elif path != testfile:
                    d.errback(Exception("path is not the original one"))
                elif newhash != testhash:
                    d.errback(Exception("the hashes are different!"))
                else:
                    d.callback(True)
        receiver = Helper()

        hasher = hash_queue._Hasher(queue, mark, receiver)
        hasher.start()

        # send what to hash
        with open(testfile, "w") as fh:
            fh.write(testinfo)
        queue.put(testfile)
        return d


class HashQueueTests(testcase.BaseTwistedTestCase):
    '''Test the whole stuff to receive signals.'''

    def setUp(self):
        '''Setup the test.'''
        testcase.BaseTwistedTestCase.setUp(self)
        self.test_dir = self.mktemp('test_dir')
        self.timeout = 2

    def tearDown(self):
        '''Clean up the tests.'''
        testcase.BaseTwistedTestCase.tearDown(self)

    def test_called_back(self):
        '''Tests that the hasher produces correct info.'''
        # create the hasher
        d = defer.Deferred()
        class Helper(object):
            '''helper class'''
            def push(self, *args):
                '''callback'''
                d.callback(args)
        receiver = Helper()
        hq = hash_queue.HashQueue(receiver)

        # send what to hash
        testfile = os.path.join(self.test_dir, "testfile")
        with open(testfile, "w") as fh:
            fh.write("foobar")
        hq.insert(testfile)

        def check_info(args):
            """check the info pushed by the hasher"""
            # pylint: disable-msg=W0612
            event, path, hash, crc, size, stat = args
            # calculate what we should receive
            realh = content_hash_factory()
            realh.hash_object.update("foobar")
            should_be = realh.content_hash()
            curr_stat = os.stat(testfile)
            self.assertEquals(should_be, hash)
            for attr in ('st_mode', 'st_ino', 'st_dev', 'st_nlink', 'st_uid',
                         'st_gid', 'st_size', 'st_ctime', 'st_mtime'):
                self.assertEquals(getattr(curr_stat, attr),
                                  getattr(stat, attr))

        d.addCallback(check_info)
        return d

    def test_order(self):
        '''The hasher should return in order.'''
        # calculate what we should receive
        should_be = []
        for i in range(10):
            hasher = content_hash_factory()
            text = "supercalifragilistico"+str(i)
            hasher.hash_object.update(text)
            tfile = os.path.join(self.test_dir, "tfile"+str(i))
            with open(tfile, "w") as fh:
                fh.write("supercalifragilistico"+str(i))
            should_be.append(("HQ_HASH_NEW", tfile, hasher.content_hash(),
                              crc32(text), len(text), os.stat(tfile)))

        d = defer.Deferred()
        class Helper(object):
            '''helper class'''
            # class-closure, cannot use self, pylint: disable-msg=E0213
            def __init__(innerself):
                innerself.store = []
            def push(innerself, *args):
                '''callback'''
                innerself.store.append(args)
                if len(innerself.store) == 10:
                    if innerself.store[:-1] == should_be[:-1]:
                        d.callback(True)
                    else:
                        d.errback(Exception("are different! "))
        receiver = Helper()

        hq = hash_queue.HashQueue(receiver)

        # send what to hash
        for i in range(10):
            tfile = os.path.join(self.test_dir, "tfile"+str(i))
            hq.insert(tfile)
        return d

    def test_unique(self):
        '''The hasher should return in order.'''
        # calculate what we should receive
        should_be = []
        for i in range(10):
            hasher = content_hash_factory()
            text = "supercalifragilistico"+str(i)
            hasher.hash_object.update(text)
            tfile = os.path.join(self.test_dir, "tfile"+str(i))
            with open(tfile, "w") as fh:
                fh.write("supercalifragilistico"+str(i))
            should_be.append(("HQ_HASH_NEW", tfile, hasher.content_hash(),
                              crc32(text), len(text), os.stat(tfile)))

        d = defer.Deferred()
        class Helper(object):
            '''helper class'''
            # class-closure, cannot use self, pylint: disable-msg=E0213
            def __init__(innerself):
                innerself.store = []
            def push(innerself, *args):
                '''callback'''
                innerself.store.append(args)
                if len(innerself.store) == 10:
                    if innerself.store[:-1] == should_be[:-1]:
                        d.callback(True)
                    else:
                        d.errback(Exception("are different! "))
        receiver = Helper()

        hq = hash_queue.HashQueue(receiver)
        # stop the hasher so we can test the unique items in the queue
        hq._queue.put(hq._end_mark)
        # create a new hasher just like the HashQueue creates it
        hasher = hash_queue._Hasher(hq._queue, hq._end_mark, receiver)
        hasher.setDaemon(True)

        # send to hash twice
        for i in range(10):
            tfile = os.path.join(self.test_dir, "tfile"+str(i))
            hq.insert(tfile)
            hq.insert(tfile)
        # start the hasher
        hasher.start()
        # insert the last items to check the unqieness of the queue while
        # the hasher is running
        for i in range(8, 10):
            tfile = os.path.join(self.test_dir, "tfile"+str(i))
            hq.insert(tfile)
        return d

    def test_interrupt_current(self):
        """ Test that the hasher correctly interrupts a inprogress task."""
        # calculate what we should receive
        testinfo = os.urandom(50000000)
        hasher = content_hash_factory()
        hasher.hash_object.update(testinfo)
        testfile = os.path.join(self.test_dir, "testfile")
        testhash = hasher.content_hash()
        # send what to hash
        with open(testfile, "w") as fh:
            fh.write(testinfo)

        d = defer.Deferred()

        class Helper(object):
            '''helper class'''
            def push(self, event, path, newhash, crc, size, stat):
                '''callback'''
                if event != "HQ_HASH_NEW":
                    d.errback(Exception("envent is not HQ_HASH_NEW"))
                elif path != testfile:
                    d.errback(Exception("path is not the original one"))
                elif newhash != testhash:
                    d.errback(Exception("the hashes are different!"))
                else:
                    d.callback(True)
        receiver = Helper()

        hq = hash_queue.HashQueue(receiver)
        hq.insert(testfile)
        # insert it again, to cancel the first one
        reactor.callLater(0.1, hq.insert, testfile)
        return d


class UniqueQueueTests(TwistedTestCase):
    """Tests for hash_queue.UniqueQueue"""

    def test_unique_elements(self):
        """Test that the queue actually holds unique elements."""
        queue = hash_queue.UniqueQueue()
        queue.put('item1')
        queue.put('item1')
        self.assertEquals(1, queue.qsize())
        self.assertEquals(1, len(queue._set))
        queue.get()
        self.assertEquals(0, queue.qsize())
        self.assertEquals(0, len(queue._set))
        queue.put('item1')
        queue.put('item2')
        queue.put('item1')
        queue.put('item2')
        self.assertEquals(2, queue.qsize())
        self.assertEquals(2, len(queue._set))
        queue.get()
        queue.get()
        self.assertEquals(0, queue.qsize())
        self.assertEquals(0, len(queue._set))


def test_suite():
    # pylint: disable-msg=C0111
    return unittest.TestLoader().loadTestsFromName(__name__)

if __name__ == "__main__":
    unittest.main()
