# ubuntuone.syncdaemon.hash_queue - hash queues
#
# Author: Facundo Batista <facundo@canonical.com>
#         Guillermo Gonzalez <guillermo.gonzalez@canonical.com>
#
# Copyright 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
'''Module that implements the Hash Queue machinery.'''

from __future__ import with_statement

import logging
import threading
import functools
import Queue
import os

from twisted.internet import reactor

from ubuntuone.storageprotocol.hash import \
    content_hash_factory, crc32


class StopHashing(Exception):
    """The current hash was cancelled"""


class _Hasher(threading.Thread):
    '''Class that lives in another thread, hashing all night long.'''
    def __init__(self, queue, end_mark, event_queue):
        self.logger = logging.getLogger('ubuntuone.SyncDaemon.HQ.hasher')
        self.end_mark = end_mark
        self.queue = queue
        self.push = functools.partial(event_queue.push, "HQ_HASH_NEW")
        self.mutex = threading.Lock()
        self._should_cancel = UniqueQueue()
        self._hashing = False # flag to indicate a hash in progress
        threading.Thread.__init__(self)


    def run(self):
        '''Run the thread.'''
        while True:
            path = self.queue.get()
            if path is self.end_mark:
                break

            self.logger.info("Hasher: got path to hash: %r" % path)
            result = self._hash(path)
            self.queue.task_done()
            if result:
                reactor.callFromThread(self.push, path, *result)
            self.logger.info("Hasher: path hash pushed:  path %r  hash %r"
                                                            % (path, result))

    def _hash(self, path):
        '''Actually hashes a file.'''
        hasher = content_hash_factory()
        crc = 0
        size = 0
        try:
            # about to start hashing 'path'
            with self.mutex:
                self._hashing = True
            initial_stat = os.stat(path)
            with open(path) as fh:
                while True:
                    path_to_cancel = self._should_cancel.get_nowait(True)
                    # stop hashing if path_to_cancel == path or the end_mark
                    if path_to_cancel == path or \
                       path_to_cancel == self.end_mark:
                        raise StopHashing('hashing of %r was cancelled' % path)
                    cont = fh.read(65536)
                    if not cont:
                        break
                    hasher.update(cont)
                    crc = crc32(cont, crc)
                    size += len(cont)
        except (IOError, OSError):
            return None
        except StopHashing, e:
            self.logger.info(str(e))
            return None
        finally:
            with self.mutex:
                self._hashing = False

        return hasher.content_hash(), crc, size, initial_stat

    def cancel_if_running(self, path):
        """request a cancel/stop of the current hash, if it's == path"""
        with self.mutex:
            if self._hashing:
                self._should_cancel.put(path)


class HashQueue(object):
    '''Interface between the real Hasher and the rest of the world.'''

    def __init__(self, event_queue):
        self.logger = logging.getLogger('ubuntuone.SyncDaemon.HQ')
        self._queue = UniqueQueue()
        self._end_mark = object()
        self.hasher = _Hasher(self._queue, self._end_mark, event_queue)
        self.hasher.setDaemon(True)
        self.hasher.start()
        self.logger.info("HashQueue: _hasher started")

    def insert(self, path):
        '''Insert the path of a file to be hashed.'''
        self.logger.info("HashQueue: inserting path %r" % path)
        self.hasher.cancel_if_running(path)
        self._queue.put(path)

    def shutdown(self):
        '''Shutdown all resources.'''
        self._queue.put(self._end_mark)
        self.logger.info("HashQueue: _hasher stopped")

    def empty(self):
        '''Return whether we are empty or not'''
        return self._queue.empty()

    def __len__(self):
        '''Return the length of the queue (not reliable!)'''
        return self._queue.qsize()


class UniqueQueue(Queue.Queue):
    """Variant of Queue that only inserts unique items in the Queue
    """

    def __init__(self, *args, **kwargs):
        """create the instance"""
        Queue.Queue.__init__(self, *args, **kwargs)
        self._set = set() # to quickly check if a item is in the queue
        self.logger = logging.getLogger('ubuntuone.SyncDaemon.HQ.Queue')

    def _put(self, item):
        """Custom _put that avoid duplicate items in the queue."""
        if item not in self._set:
            Queue.Queue._put(self, item)
            self._set.add(item)
        else:
            self.logger.debug('Item already in the queue: %r', item)

    def _get(self):
        """Custom _get that removes the item from self._set."""
        item = Queue.Queue._get(self)
        self._set.discard(item)
        return item

    def get_nowait(self, none_on_empty=False):
        """ custom version that returns None if the Queue is empty and
        accepts none_on_empty kwarg (defaults to False).

        If none_on_empty=True, returns None instead of raising Empty.
        """
        if none_on_empty:
            try:
                return Queue.Queue.get_nowait(self)
            except Queue.Empty:
                return None
        return Queue.Queue.get_nowait(self)

